package org.ids_mannheim;

import org.anarres.parallelgzip.ParallelGZIPOutputStream;
import org.apache.commons.compress.compressors.xz.XZCompressorOutputStream;
import picocli.CommandLine;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Locale;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

@CommandLine.Command(mixinStandardHelpOptions = true, name = "totalngrams", version = "2.1.0", description = "sum ngram counts from KorAP-XML, CoNLL-U files and frequency lists")

public class TotalNGrams implements Callable<Integer> {
    static public final Logger logger;
    private static final int MAX_THREADS = Runtime.getRuntime().availableProcessors() - 2;
    private static final String DEFAULT_LOGFILE = "totalngrams.log";

    static {
        String path = TotalNGrams.class.getClassLoader()
                .getResource("logging.properties")
                .getFile();
        System.setProperty("java.util.logging.config.file", path);
        logger = Logger.getLogger(TotalNGrams.class.getSimpleName());
    }

    @CommandLine.Parameters(arity = "1..*", description = "input files")
    private ArrayList<String> inputFiles = new ArrayList<>();
    @CommandLine.Option(names = { "-L",
            "--log-file" }, defaultValue = DEFAULT_LOGFILE, description = "log file name (default: ${DEFAULT-VALUE})")
    String logFileName;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-o",
            "--output-file" }, description = "Output file (default: ${DEFAULT-VALUE})")
    String output_fillename = "-";

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-d",
            "--downcase" }, description = "Convert all token characters into lower case (default: ${DEFAULT-VALUE})")
    boolean downcase_tokens = false;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "--force" }, description = "Force overwrite (default: ${DEFAULT-VALUE})")
    boolean force_overwrite = false;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-l",
            "--with-lemma-pos" }, description = "Use also lemma and part-of-speech annotations (default: ${DEFAULT-VALUE}")
    boolean with_lemma_and_pos = false;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-P",
            "--max-procs" }, description = "Run up to max-procs processes at a time (default: ${DEFAULT-VALUE})")
    int max_threads = MAX_THREADS;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-p",
            "--worker-pool" }, description = "Run preprocessing on extern hosts, e.g. '10*local,5*host1,3*smith@host2' (default: ${DEFAULT-VALUE})")
    String worker_pool_specification = "";

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-n",
            "--ngram-size" }, description = "n-gram size (default: ${DEFAULT-VALUE})")
    int ngram_size = 1;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-F",
            "--folds" }, description = "number of random folds (default: ${DEFAULT-VALUE})")
    int FOLDS = 1;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-f",
            "--fold" }, description = "current fold (default: ${DEFAULT-VALUE})")
    int fold = 1;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-S",
            "--sort" }, description = "Toggle output sorting (default: ${DEFAULT-VALUE})")
    boolean sort = true;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = { "-N",
            "--numeric-secondary-sort" }, description = "Sort entries with same frequency numerically  (default: ${DEFAULT-VALUE})")
    boolean numericSecondarySort = false;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {
            "--pad" }, description = "Add padding " + PaddedSlidingWindowQueue.TEXT_START_SYMBOL + " and "
                    + PaddedSlidingWindowQueue.TEXT_END_SYMBOL + " symbols at text edges  (default: ${DEFAULT-VALUE})")
    boolean addPadding = false;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {
            "--exclude-punctuation" }, description = "Ignore all tokens tagged as punctuation (according to STTS tags set, i.e. starting with '$') (default: ${DEFAULT-VALUE})")
    boolean excludePunctuation = false;

    private Progressbar etaPrinter;

    public TotalNGrams() {
    }

    public static void main(String[] args) {
        System.exit(new CommandLine(new TotalNGrams()).execute(args));
    }

    private void initLogging() {
        if (logFileName.equals(DEFAULT_LOGFILE) && !output_fillename.equals("-")) {
            logFileName = Utils.newExtension(output_fillename, "log");
        }
        try {
            FileHandler fileHandler = new FileHandler(logFileName);
            System.setProperty("java.util.logging.SimpleFormatter.format",
                    "[%1$tF %1$tT] [%4$-7s] %5$s %n");
            fileHandler.setFormatter(new SimpleFormatter());
            // java.util.logging.SimpleFormatter.format="%1$tb %1$td, %1$tY
            // %1$tl:%1$tM:%1$tS %1$Tp %2$s%n%4$s: %5$s%n";
            java.util.Locale.setDefault(Locale.Category.FORMAT, Locale.ROOT);
            logger.addHandler(fileHandler);
            ConsoleHandler consoleHandler = new ConsoleHandler();
            consoleHandler.setLevel(Level.WARNING);
            logger.addHandler(consoleHandler);
        } catch (Exception e) {
            System.err.println("Cannot write logfile " + logFileName + ": " + e);
            System.exit(-1);
        }
    }

    @Override
    public Integer call() throws Exception {
        initLogging();
        final PrintStream output_stream;
        if ((output_fillename == null) || output_fillename.equals("-")) {
            output_stream = System.out;
        } else {
            File f = Utils.createFile(output_fillename, force_overwrite);
            if (output_fillename.endsWith(".gz")) {
                output_stream = new PrintStream(
                        new ParallelGZIPOutputStream((new BufferedOutputStream(new FileOutputStream(f)))));
            } else if (output_fillename.endsWith(".xz")) {
                output_stream = new PrintStream(
                        new XZCompressorOutputStream((new BufferedOutputStream(new FileOutputStream(f)))));
            } else {
                output_stream = new PrintStream(new BufferedOutputStream(new FileOutputStream(f)));
            }

        }

        ConcurrentHashMap<String, AtomicInteger> map = new ConcurrentHashMap<>();

        long totalFilesSizes = inputFiles.parallelStream().mapToLong(fname -> {
            long l = new File(fname).length();
            if (l == 0) {
                logger.severe("File " + fname + " does not exist or is empty.");
                System.exit(-1);
            }
            return l;
        }).sum();
        inputFiles = (ArrayList<String>) inputFiles.parallelStream()
                .sorted((a, b) -> Long.compareUnsigned(new File(b).length(), new File(a).length()))
                .collect(Collectors.toList());

        etaPrinter = new Progressbar(totalFilesSizes);
        BlockingQueue<Integer> queue = new LinkedBlockingQueue<>(inputFiles.size());
        ExecutorService es = Executors.newCachedThreadPool();
        WorkerNodePool workerNodePool = new WorkerNodePool(worker_pool_specification);
        if (!worker_pool_specification.equals("")) {
            max_threads = workerNodePool.size;
        }
        int threads = Math.min(max_threads, inputFiles.size());
        logger.info("Processing fold " + fold + "/" + FOLDS);
        logger.info("Using " + threads + " threads");
        IntStream.range(0, threads).forEach(unused -> es.execute(new Worker(queue, inputFiles, ngram_size, fold, FOLDS,
                map, with_lemma_and_pos, downcase_tokens, workerNodePool, etaPrinter, logger, addPadding, excludePunctuation)));
        queue.addAll(IntStream.range(0, inputFiles.size()).boxed().collect(Collectors.toList()));
        IntStream.range(0, threads).forEach(unused -> {
            try {
                queue.put(-1);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        });
        es.shutdown();
        boolean finished = es.awaitTermination(120, TimeUnit.HOURS);
        etaPrinter.finish();
        if (!sort) {
            logger.info("Writing unsorted frequency table.");
            System.err.println("Writing unsorted frequency table.");
            map.forEach((key, value) -> output_stream.println(key +
                    '\t' +
                    Integer.toUnsignedString(value.get())));
        } else {
            logger.info("Sorting and writing frequency table.");
            System.err.println("Sorting and writing frequency table.");
            map.entrySet().parallelStream()
                    .sorted((a, b) -> {
                        int cmp1 = Integer.compareUnsigned(b.getValue().get(), a.getValue().get());
                        if (cmp1 != 0) {
                            return cmp1;
                        } else {
                            if (numericSecondarySort) {
                                String[] columnsA = a.getKey().split("\t");
                                String[] columnsB = b.getKey().split("\t");
                                int len = Math.min(columnsA.length, columnsB.length);
                                for (int i = 0; i < len; i++) {
                                    int res = new BigInteger(columnsA[i]).compareTo(new BigInteger(columnsB[i]));
                                    if (res != 0)
                                        return (res);
                                }
                                return 0;
                            } else
                                return a.getKey().compareTo(b.getKey());
                        }
                    })
                    .forEachOrdered(entry -> output_stream
                            .println(entry.getKey() + "\t" + Integer.toUnsignedString(entry.getValue().get())));
        }
        logger.info("Finished.");
        output_stream.close();
        return null;
    }
}
